/* 
 *  This program is an interactive FAQ to ease investigations.
 *  Copyright (C) 2012 Julien Bruguier.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */ 

#ifndef __INVESTIGATION_DONNEES_AUTOMATE__
#define __INVESTIGATION_DONNEES_AUTOMATE__

#include <set>
#include <vector>
#include <map>
#include <src/global/global.h>
namespace Investigation { namespace Donnees { } }
#include <src/global/exceptions.h>
#include <src/global/sources/position.h>

namespace Investigation
{
	namespace Donnees
	{
		template<typename Etat, typename Lettre>
		struct BoiteOutils
		{
			typedef SHARED_PTR(Etat) EtatSP;
			typedef WEAK_PTR(Etat) EtatWP;
			struct Compare
			{
				bool operator()(const EtatSP& g, const EtatSP& d) const
				{
					return (*g)<(*d);
				}
			};
			typedef std::set<EtatSP, Compare > EnsembleDEtats;
			typedef std::map<EtatSP, EnsembleDEtats, Compare> AssociationDEtats;
			typedef std::set<Lettre> Alphabet;
			typedef std::vector<Lettre> Mot;
		};

		template<template <typename Etat, typename Lettre> class Type, typename Etat, typename Lettre>
			struct Transitions
			{
				Transitions(const typename BoiteOutils<Etat,Lettre>::Alphabet& alphabet)
				{
					for(auto it=alphabet.cbegin() ; it!=alphabet.cend() ; ++it)
					{
						_transitions_constantes[*it] = typename Type<Etat,Lettre>::Destination();
					}
				};
				std::map<Lettre, typename Type<Etat,Lettre>::Destination> _transitions_constantes;
				typename Type<Etat,Lettre>::Destination _transition_autres;
				typename Type<Etat,Lettre>::Destination _transition_instantanee;
			};

		template<typename Etat, typename Lettre>
			struct Deterministe
			{
				typedef typename BoiteOutils<Etat,Lettre>::EtatSP Destination;
				typedef typename BoiteOutils<Etat,Lettre>::EtatSP Initial;
				static void ajout_destination(Destination& destination, const Destination& etat)
				{
					if(destination) throw;
					destination = etat;
				}
				static void ajout_initial(Initial& initial, const Initial& etat)
				{
					if(initial) throw;
					initial = etat;
				}
				static typename BoiteOutils<Etat,Lettre>::EnsembleDEtats etats(const Initial& etat)
				{
					typename BoiteOutils<Etat,Lettre>::EnsembleDEtats etats;
					if(etat)
					{
						etats.insert(etat);
					}
					return etats;
				}
				static Destination lecture(const Lettre& lettre, const Destination& etat_courant, const std::map<typename BoiteOutils<Etat,Lettre>::EtatSP,Transitions<Deterministe,Etat,Lettre>,typename BoiteOutils<Etat,Lettre>::Compare >& transitions)
				{
					auto ittransitions = transitions.find(etat_courant);
					if(ittransitions==transitions.end())
						throw;
					auto itconst = ittransitions->second._transitions_constantes.find(lettre);
					if(itconst!=ittransitions->second._transitions_constantes.end())
					{
						return itconst->second;
					}
					return ittransitions->second._transition_autres;
				}
				static bool etat_valide(const Destination& etat)
				{
					return static_cast<bool>(etat);
				}
				static bool etat_final(const Destination& etat, const typename BoiteOutils<Etat,Lettre>::EnsembleDEtats& finaux)
				{
					return (finaux.find(etat)!=finaux.end());
				}
				static std::string affiche_initial(const std::map<typename BoiteOutils<Etat,Lettre>::EtatSP, size_t>& etats, const Initial& etat)
				{
					std::ostringstream oss;
					if(etat)
					{
						oss << etats.at(etat);
					}
					return oss.str();
				}
				static std::string affiche_destination(const std::map<typename BoiteOutils<Etat,Lettre>::EtatSP, size_t>& etats, const Destination& etat)
				{
					std::ostringstream oss;
					if(etat)
					{
						oss << etats.at(etat);
					}
					return oss.str();
				}
			};

		template<typename Etat, typename Lettre>
			struct NonDeterministe
			{
				typedef typename BoiteOutils<Etat,Lettre>::EnsembleDEtats Destination;
				typedef typename BoiteOutils<Etat,Lettre>::EnsembleDEtats Initial;
				static void ajout_destination(Destination& destination, const SHARED_PTR(Etat)& etat)
				{
					destination.insert(etat);
				}
				static void ajout_destination(Destination& destination, const Destination& etat)
				{
					destination.insert(etat.begin(),etat.end());
				}
				static void ajout_initial(Initial& initial, const SHARED_PTR(Etat)& etat)
				{
					initial.insert(etat);
				}
				static typename BoiteOutils<Etat,Lettre>::EnsembleDEtats etats(const Initial& etat)
				{
					return etat;
				}
				static Destination lecture(const Lettre& lettre, const Destination& etat_courant, const std::map<typename BoiteOutils<Etat,Lettre>::EtatSP,Transitions<NonDeterministe,Etat,Lettre>,typename BoiteOutils<Etat,Lettre>::Compare >& transitions)
				{
					throw;
				}
				static bool etat_valide(const Destination& etat)
				{
					return not etat.empty();
				}
				static bool etat_final(const Destination& etat, const typename BoiteOutils<Etat,Lettre>::EnsembleDEtats& finaux)
				{
					for(auto it=etat.cbegin() ; it!=etat.cend() ; ++it)
					{
						if(finaux.find(*it)!=finaux.end())
						{
							return true;
						}
					}
					return false;
				}
				static std::string affiche_initial(const std::map<typename BoiteOutils<Etat,Lettre>::EtatSP, size_t>& etats, const Initial& etat)
				{
					std::ostringstream oss;
					for(auto it=etat.cbegin() ; it!=etat.cend() ; ++it)
					{
						if(it!=etat.cbegin())
						{
							oss << ", ";
						}
						oss << etats.at(*it);
					}
					return oss.str();
				}
				static std::string affiche_destination(const std::map<typename BoiteOutils<Etat,Lettre>::EtatSP, size_t>& etats, const Destination& etat)
				{
					std::ostringstream oss;
					for(auto it=etat.cbegin() ; it!=etat.cend() ; ++it)
					{
						if(it!=etat.cbegin())
						{
							oss << ", ";
						}
						oss << etats.at(*it);
					}
					return oss.str();
				}
			};
	
		template<typename Etat, typename Lettre>
			struct NonDeterministeSimplifie : public NonDeterministe<Etat,Lettre>
			{
				static typename NonDeterministe<Etat,Lettre>::Destination lecture(const Lettre& lettre, const typename NonDeterministe<Etat,Lettre>::Destination& etat_courant, const std::map<typename BoiteOutils<Etat,Lettre>::EtatSP,Transitions<NonDeterministeSimplifie,Etat,Lettre>,typename BoiteOutils<Etat,Lettre>::Compare >& transitions)
				{
					typename NonDeterministe<Etat,Lettre>::Destination resultat;
					for(auto it=etat_courant.cbegin() ; it!=etat_courant.cend() ; ++it)
					{
						auto ittransitions = transitions.find(*it);
						if(ittransitions==transitions.end())
							throw;
						auto itconst = ittransitions->second._transitions_constantes.find(lettre);
						if(itconst!=ittransitions->second._transitions_constantes.end())
						{
							resultat.insert(itconst->second.cbegin(),itconst->second.cend());
						}
						else
						{
							resultat.insert(ittransitions->second._transition_autres.cbegin(),ittransitions->second._transition_autres.cend());
						}
					}
					return resultat;
				}
			};

		struct EtatSimple
		{
			bool operator<(const EtatSimple& etat)
			{
				return this<(&etat);
			}
		};

		template<typename Etat, typename Lettre>
		struct EtatDeterminise
		{
			public:
				EtatDeterminise(const typename BoiteOutils<Etat,Lettre>::EnsembleDEtats& etats)
				:_etats(etats) { };
				bool operator< (const EtatDeterminise& etat) const
				{
					if(_etats.size()<etat._etats.size())
						return true;
					if(etat._etats.size()<_etats.size())
						return false;
					for(auto itg=_etats.cbegin(), itd=etat._etats.cbegin() ; itg!=_etats.cend() ; ++itg, ++itd)
					{
						if((**itg)<(**itd))
							return true;
						if((**itd)<(**itg))
							return false;
					}
					return false;
				}
				typename BoiteOutils<Etat,Lettre>::EnsembleDEtats::const_iterator cbegin() const
				{
					return _etats.cbegin();
				}
				typename BoiteOutils<Etat,Lettre>::EnsembleDEtats::const_iterator cend() const
				{
					return _etats.cend();
				}
			private:
				typename BoiteOutils<Etat,Lettre>::EnsembleDEtats _etats;
		};

		template<template<typename Etat, typename Lettre> class Type, typename Lettre, typename Etat = EtatSimple>
			struct Automate
			{
				explicit Automate(const typename BoiteOutils<Etat,Lettre>::Alphabet& alphabet)
					:_alphabet(alphabet) {};

				template<typename EtatAD>
					static SHARED_PTR(EtatAD) etat_correspondant(std::map<EtatDeterminise<Etat,Lettre> , SHARED_PTR(EtatAD) >& etats, const EtatDeterminise<Etat,Lettre>& etat)
					{
						auto it=etats.find(etat);
						if(it!=etats.end())
						{
							return it->second;
						}
						return etats[etat] = SHARED_PTR(EtatAD)(new EtatAD());
					};

				bool operator() (const typename BoiteOutils<Etat,Lettre>::Mot& mot) const
				{
					typename Type<Etat,Lettre>::Destination etat_courant = _etats_initiaux;
					for(auto itmot=mot.cbegin() ; itmot!=mot.cend() ; ++itmot)
					{
						etat_courant = Type<Etat,Lettre>::lecture(*itmot,etat_courant,_transitions);
						if(not Type<Etat,Lettre>::etat_valide(etat_courant))
							return false;
					}
					return Type<Etat,Lettre>::etat_final(etat_courant,_etats_finaux);
				};

				operator Automate<NonDeterministeSimplifie,Lettre,Etat> () const
				{
					Automate<NonDeterministeSimplifie,Lettre,Etat> automate(_alphabet);
					typename BoiteOutils<Etat,Lettre>::AssociationDEtats transitions_instantanees_inverses;
					for(auto it=_transitions.cbegin() ; it!=_transitions.cend() ; ++it)
					{
						typename BoiteOutils<Etat,Lettre>::EnsembleDEtats etats;
						etats.insert(it->first);
						etats = accessibles_instantanement(etats);
						for(auto itit=etats.cbegin() ; itit!=etats.cend() ; ++itit)
						{
							transitions_instantanees_inverses[*itit].insert(it->first);
						}
					}
					automate._etats_initiaux = _etats_initiaux;
					Transitions<NonDeterministeSimplifie,Etat,Lettre> transitions(_alphabet);
					for(auto it=_transitions.cbegin() ; it!=_transitions.cend() ; ++it)
					{
						automate._transitions.insert(std::pair<typename BoiteOutils<Etat,Lettre>::EtatSP, Transitions<NonDeterministeSimplifie,Etat,Lettre> >(it->first, transitions));
					}
					for(auto itetat=_transitions.cbegin() ; itetat!=_transitions.cend() ; ++itetat)
					{
						for(auto itlettre=itetat->second._transitions_constantes.cbegin() ; itlettre!=itetat->second._transitions_constantes.cend() ; ++itlettre)
						{
							for(auto itdestination=itlettre->second.cbegin() ; itdestination!=itlettre->second.cend() ; ++itdestination)
							{
								for(auto itorigine=transitions_instantanees_inverses[itetat->first].cbegin() ; itorigine!=transitions_instantanees_inverses[itetat->first].cend() ; ++itorigine)
								{
									automate.ajout_transition(*itorigine,itlettre->first,*itdestination);
								}
							}
						}
						for(auto itdestination=itetat->second._transition_autres.cbegin() ; itdestination!=itetat->second._transition_autres.cend() ; ++itdestination)
						{
							for(auto itorigine=transitions_instantanees_inverses[itetat->first].cbegin() ; itorigine!=transitions_instantanees_inverses[itetat->first].cend() ; ++itorigine)
							{
								automate.ajout_transition(*itorigine,*itdestination);
							}
						}
					}
					for(auto it=_etats_finaux.cbegin() ; it!=_etats_finaux.cend() ; ++it)
					{
						automate._etats_finaux.insert(transitions_instantanees_inverses[*it].cbegin(),transitions_instantanees_inverses[*it].cend());
					}
					automate.emonde();
					return automate;
				};

				template<typename EtatAD>
					operator Automate<Deterministe,Lettre,EtatAD> () const
					{
						Automate<Deterministe,Lettre,EtatAD> automate(_alphabet);
						std::map<EtatDeterminise<Etat,Lettre> , SHARED_PTR(EtatAD) > etats;
						std::vector<EtatDeterminise<Etat,Lettre> > pile_d_etats;
						auto etats_initiaux = accessibles_instantanement(_etats_initiaux); 
						automate._etats_initiaux = etat_correspondant(etats,etats_initiaux);
						pile_d_etats.push_back(etats_initiaux);
						while(not pile_d_etats.empty())
						{
							auto etat = pile_d_etats.back();
							pile_d_etats.pop_back();
							auto etat_equivalent = etat_correspondant(etats,etat);
							if(automate._transitions.find(etat_equivalent)!=automate._transitions.end())
								continue;
							Transitions<Deterministe,EtatAD,Lettre> transitions(_alphabet);
							for(auto itetat=etat.cbegin() ; itetat!=etat.cend() ; ++itetat)
							{
								for(auto italpha = _alphabet.cbegin() ; italpha!=_alphabet.cend() ; ++italpha)
								{
									auto etats_atteints = Type<Etat,Lettre>::etats(_transitions.at(*itetat)._transitions_constantes.at(*italpha));
									if(not etats_atteints.empty())
									{
										etats_atteints = accessibles_instantanement(etats_atteints);
										transitions._transitions_constantes[*italpha]=etat_correspondant(etats,etats_atteints);
										pile_d_etats.push_back(etats_atteints);
									}
								}
								{
									auto etats_atteints = Type<Etat,Lettre>::etats(_transitions.at(*itetat)._transition_autres);
									if(not etats_atteints.empty())
									{
										etats_atteints = accessibles_instantanement(etats_atteints);
										transitions._transition_autres=etat_correspondant(etats,etats_atteints);
										pile_d_etats.push_back(etats_atteints);
									}
								}
							}
							automate._transitions.insert(std::pair<SHARED_PTR(EtatAD),Transitions<Deterministe,EtatAD,Lettre> >(etat_equivalent,transitions));
						}
						for(auto itetats = etats.cbegin() ; itetats!=etats.cend() ; ++itetats)
						{
							for(auto it=itetats->first.cbegin() ; it!=itetats->first.cend(); ++it)
							{
								if(_etats_finaux.find(*it)!=_etats_finaux.end())
								{
									automate._etats_finaux.insert(itetats->second);
								}
							}
						}
						automate.emonde();
						return automate;
					};

				typename BoiteOutils<Etat,Lettre>::EnsembleDEtats accessibles_instantanement(const typename BoiteOutils<Etat,Lettre>::EnsembleDEtats& origines) const
				{
					typename BoiteOutils<Etat,Lettre>::EnsembleDEtats resultat = origines;
					for(auto it=origines.cbegin() ; it!=origines.cend() ; ++it)
					{
						resultat.insert(_transitions.at(*it)._transition_instantanee.begin(),_transitions.at(*it)._transition_instantanee.end());
					}
					if(resultat.size()==origines.size())
						return resultat;
					return accessibles_instantanement(resultat);
				}

				void emonde()
				{
					for(;;)
					{
						typename BoiteOutils<Etat,Lettre>::EnsembleDEtats etats_accessibles = Type<Etat,Lettre>::etats(_etats_initiaux);
						for(auto itetat=_transitions.cbegin() ; itetat!=_transitions.cend() ; ++itetat)
						{
							typename BoiteOutils<Etat,Lettre>::EnsembleDEtats etats;
							for(auto itlettre=itetat->second._transitions_constantes.cbegin() ; itlettre!=itetat->second._transitions_constantes.cend() ; ++itlettre)
							{
								etats = Type<Etat,Lettre>::etats(itlettre->second);
								etats_accessibles.insert(etats.cbegin(),etats.cend());
							}
							etats = Type<Etat,Lettre>::etats(itetat->second._transition_autres);
							etats_accessibles.insert(etats.cbegin(),etats.cend());
							etats = Type<Etat,Lettre>::etats(itetat->second._transition_instantanee);
							etats_accessibles.insert(etats.cbegin(),etats.cend());
						}
						if(etats_accessibles.size()==_transitions.size())
							return;
						std::map<typename BoiteOutils<Etat,Lettre>::EtatSP,Transitions<Type,Etat,Lettre>,typename BoiteOutils<Etat,Lettre>::Compare > transitions;
						typename BoiteOutils<Etat,Lettre>::EnsembleDEtats etats_finaux;
						for(auto itetat=_transitions.cbegin() ; itetat!=_transitions.cend() ; ++itetat)
						{
							if(etats_accessibles.find(itetat->first)!=etats_accessibles.end())
							{
								transitions.insert(*itetat);
								if(_etats_finaux.find(itetat->first)!=_etats_finaux.end())
								{
									etats_finaux.insert(itetat->first);
								}
							}
						}
						_transitions = transitions;
						_etats_finaux = etats_finaux;
					}
					
				}

				template<typename... Args>
					typename BoiteOutils<Etat,Lettre>::EtatSP ajout_etat(Args... args)
					{
						typename BoiteOutils<Etat,Lettre>::EtatSP nouvel_etat(new Etat(args...));
						_transitions.insert(std::pair<typename BoiteOutils<Etat,Lettre>::EtatSP,Transitions<Type,Etat,Lettre> >(nouvel_etat,_alphabet));
						return nouvel_etat;
					};
				template<typename... Args>
					typename BoiteOutils<Etat,Lettre>::EtatSP ajout_etat_initial(Args... args)
					{
						auto nouvel_etat = ajout_etat(args...);
						Type<Etat,Lettre>::ajout_initial(_etats_initiaux,nouvel_etat);
						return nouvel_etat;
					};
				template<typename... Args>
					typename BoiteOutils<Etat,Lettre>::EtatSP ajout_etat_final(Args... args)
					{
						auto nouvel_etat = ajout_etat(args...);
						_etats_finaux.insert(nouvel_etat);
						return nouvel_etat;
					};
				void ajout_transition(const typename BoiteOutils<Etat,Lettre>::EtatSP& origine, const Lettre& lettre, const typename BoiteOutils<Etat,Lettre>::EtatSP& destination)
				{
					if(_alphabet.find(lettre)==_alphabet.end()) throw;
					auto it = _transitions.find(origine);
					if(it==_transitions.end()) throw;
					Type<Etat,Lettre>::ajout_destination(it->second._transitions_constantes[lettre],destination);
				};
				void ajout_transition(const typename BoiteOutils<Etat,Lettre>::EtatSP& origine, const typename BoiteOutils<Etat,Lettre>::EtatSP& destination)
				{
					auto it = _transitions.find(origine);
					if(it==_transitions.end()) throw;
					Type<Etat,Lettre>::ajout_destination(it->second._transition_autres,destination);
				};
				void ajout_transition_instantanee(const typename BoiteOutils<Etat,Lettre>::EtatSP& origine, const typename Type<Etat,Lettre>::Destination& destination)
				{
					auto it = _transitions.find(origine);
					if(it==_transitions.end()) throw;
					Type<Etat,Lettre>::ajout_destination(it->second._transition_instantanee,destination);
				};
				void ajout_transition_instantanee(const typename BoiteOutils<Etat,Lettre>::EnsembleDEtats& origine, const typename Type<Etat,Lettre>::Destination& destination)
				{
					for(auto it = origine.cbegin() ; it!=origine.cend() ; ++it)
					{
						ajout_transition_instantanee(*it,destination);
					}
				};
				void ajout_automate(const Automate<Type,Lettre,Etat>& automate)
				{
					if(_alphabet!=automate._alphabet) throw;
					_transitions.insert(automate._transitions.begin(),automate._transitions.end());
				};
				std::map<typename BoiteOutils<Etat,Lettre>::EtatSP,Transitions<Type,Etat,Lettre>,typename BoiteOutils<Etat,Lettre>::Compare > _transitions;
				typename BoiteOutils<Etat,Lettre>::Alphabet _alphabet;
				typename Type<Etat,Lettre>::Initial _etats_initiaux;
				typename BoiteOutils<Etat,Lettre>::EnsembleDEtats _etats_finaux;

				template<typename oStream>
					friend oStream& operator<<(oStream& os, const Automate<Type,Lettre,Etat>& automate)
					{
						std::map<typename BoiteOutils<Etat,Lettre>::EtatSP, size_t> etats;
						size_t index = 0;
						for(auto it=automate._transitions.cbegin() ; it!=automate._transitions.cend() ; ++it)
						{
							etats[it->first] = ++index;
						}
						os << "FSM:" << std::endl;
						os << "Initial states: ";
						os << Type<Etat,Lettre>::affiche_initial(etats,automate._etats_initiaux);
						os << std::endl;
						os << "Final states  : ";
						for(auto it=automate._etats_finaux.cbegin() ; it!=automate._etats_finaux.cend() ; ++it)
						{
							if(it!=automate._etats_finaux.cbegin())
							{
								os << ", ";
							}
							os << etats[*it];
						}
						os << std::endl;
						os << "Transitions:" << std::endl;
						std::vector<std::vector<std::string> > transitions;
						std::vector<std::string> ligne;
						ligne.push_back("");
						for(auto it=etats.begin() ; it!=etats.end() ; ++it)
						{
							std::ostringstream oss;
							oss << it->second;
							ligne.push_back(oss.str());
						}
						transitions.push_back(ligne);
						for(auto itlettre=automate._alphabet.cbegin() ; itlettre!=automate._alphabet.cend() ; ++itlettre)
						{
							ligne.clear();
							ligne.push_back(*itlettre);
							for(auto itetat=etats.cbegin() ; itetat!=etats.cend() ; ++itetat)
							{
								ligne.push_back(Type<Etat,Lettre>::affiche_destination(etats,automate._transitions.at(itetat->first)._transitions_constantes.at(*itlettre)));
							}
							transitions.push_back(ligne);
						}
						ligne.clear();
						ligne.push_back("*");
						for(auto itetat=etats.cbegin() ; itetat!=etats.cend() ; ++itetat)
						{
							ligne.push_back(Type<Etat,Lettre>::affiche_destination(etats,automate._transitions.at(itetat->first)._transition_autres));
						}
						transitions.push_back(ligne);
						ligne.clear();
						ligne.push_back("");
						for(auto itetat=etats.cbegin() ; itetat!=etats.cend() ; ++itetat)
						{
							ligne.push_back(Type<Etat,Lettre>::affiche_destination(etats,automate._transitions.at(itetat->first)._transition_instantanee));
						}
						transitions.push_back(ligne);
						index=0;
						for(auto itlettre=transitions.begin() ; itlettre!=transitions.end() ; ++itlettre)
						{
							for(auto itetat=itlettre->begin() ; itetat!=itlettre->end() ; ++itetat)
							{
								if(itetat->size()>index)
									index=itetat->size();
							}
						}
						for(auto itlettre=transitions.begin() ; itlettre!=transitions.end() ; ++itlettre)
						{
							for(auto itetat=itlettre->begin() ; itetat!=itlettre->end() ; ++itetat)
							{
								std::ostringstream oss;
								for(auto it = itetat->size() ; it<index ; ++it)
								{
									oss << " ";
								}
								oss << *itetat;
								(*itetat) = oss.str();
							}
						}
						for(auto itlettre=transitions.cbegin() ; itlettre!=transitions.cend() ; ++itlettre)
						{
							os << std::endl;
							for(auto itetat=itlettre->begin() ; itetat!=itlettre->end() ; ++itetat)
							{
								os << (*itetat) << " | ";
							}
							os << std::endl;
						}

						os << "End of FSM" << std::endl << std::endl;
						return os;
					}

			};
	}
}
#endif
